------------------------------------------------------------------------------
--                         ADAGIO - ADALID - AENEA.                         --
--                                                                          --
--                            Copyright (C) 2003                            --
--                                 A. Mosteo.                               --
--                                                                          --
--  Authors: A. Mosteo. (adagio@mosteo.com)                                 --
--                                                                          --
--  If you have any questions in regard to this software, please address    --
--  them to the above email.                                                --
--                                                                          --
--  This program is free software; you can redistribute it and/or modify    --
--  it under the terms of the GNU General Public License as published by    --
--  the Free Software Foundation; either version 2 of the License, or (at   --
--  your option) any later version.                                         --
--                                                                          --
--  This program is distributed in the hope that it will be useful, but     --
--  WITHOUT ANY WARRANTY; without even the implied warranty of              --
--  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU       --
--  General Public License for more details.                                --
--                                                                          --
--  You should have received a copy of the GNU General Public License       --
--  along with this library; if not, write to the Free Software Foundation, --
--  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.          --
--                                                                          --
--  You are not allowed to use any part of this code to develop a program   --
--  whose output would be used to harass or prosecute other users of the    --
--  networks Adagio connects with. All data collected with Adagio or a tool --
--  containing Adagio code about other network users must remain            --
--  confidential and cannot be made public by any mean, nor be used to      --
--  harass or legally prosecute these users.                                --
------------------------------------------------------------------------------
--  $Id: aenea-hub.ads,v 1.11 2004/03/22 07:14:55 Jano Exp $

with Charles.Hash_string;
with Charles.Maps.Hashed.Strings.Unbounded;

with Agpl.Geoip;

with Ada.Calendar;
use  Ada;

package Aenea.Hub is

   ------------------------------------------------------------------------
   -- Object                                                             --
   ------------------------------------------------------------------------
   -- A single server found
   type Object is private;
   type Object_access is access all Object;

   ------------------------------------------------------------------------
   -- Links                                                              --
   ------------------------------------------------------------------------
   -- This model a relationship of peer between two hubs.
   type Links is record
      Peer  : Object_access; -- The destiny of the link
      Since : Calendar.Time; -- Since when this link is established
   end record;

   -- Only takes into account the id of the destination peer:
   function Same_link (L, R : in Links) return Boolean;
   pragma Inline (Same_link);

   ------------------------------------------------------------------------
   -- Hub_map                                                            --
   ------------------------------------------------------------------------
   package Hub_map is new Charles.Maps.Hashed.Strings.Unbounded (
      Object_access,
      Charles.Hash_string,
      "=",
      "=");

   package Links_map is new Charles.Maps.Hashed.Strings.Unbounded (
      Links,
      Charles.Hash_string,
      "=",
      "=");

   ------------------------------------------------------------------------
   -- Create                                                             --
   ------------------------------------------------------------------------
   procedure Create (
      This    : out Object; 
      Address : in String; 
      Leaves  : in Natural := 0;
      Alive   : in Boolean := false);

   ------------------------------------------------------------------------
   -- Id                                                                 --
   ------------------------------------------------------------------------
   -- Returns a string uniquely identifying the hub (ip:port)
   function Id (This : in Object) return String;
   pragma Inline (Id);
   
   ------------------------------------------------------------------------
   -- Ping                                                               --
   ------------------------------------------------------------------------
   -- Launchs a ping
   procedure Ping (This : in out Object);

   ------------------------------------------------------------------------
   -- Query                                                              --
   ------------------------------------------------------------------------
   -- Launchs a query
   -- If not always, will fail for a dead hub
   procedure Query (This : in out Object);

   ------------------------------------------------------------------------
   -- Kill                                                               --
   ------------------------------------------------------------------------
   -- Mark a hub as dead (or not hub)
   procedure Kill (This : in out Object);

   ------------------------------------------------------------------------
   -- Set_alive                                                          --
   ------------------------------------------------------------------------
   procedure Set_alive (This : in out Object; Alive : in Boolean := true);

   ------------------------------------------------------------------------
   -- Set_leaves                                                         --
   ------------------------------------------------------------------------
   procedure Set_leaves (This : in out Object; Leaves : Natural);

   ------------------------------------------------------------------------
   -- Leaves                                                             --
   ------------------------------------------------------------------------
   function Leaves (This : in Object) return Natural;
      
   ------------------------------------------------------------------------
   -- Set_max_leaves                                                     --
   ------------------------------------------------------------------------
   procedure Set_max_leaves (This : in out Object; Leaves : Natural);

   ------------------------------------------------------------------------
   -- Max_leaves                                                         --
   ------------------------------------------------------------------------
   function Max_leaves (This : in Object) return Natural;

   ------------------------------------------------------------------------
   -- Alive                                                              --
   ------------------------------------------------------------------------
   function Alive (This : in Object) return Boolean;

   ------------------------------------------------------------------------
   -- Set_answer_pending                                                 --
   ------------------------------------------------------------------------
   procedure Set_answer_pending (
      This : in out Object; Answer_pending : in Boolean := true);

   ------------------------------------------------------------------------
   -- Answer_pending                                                     --
   ------------------------------------------------------------------------
   function Answer_pending (This : in Object) return Boolean;

   ------------------------------------------------------------------------
   -- Add_answer                                                         --
   ------------------------------------------------------------------------
   -- Adds 1 to the answers statistic
   procedure Add_answer (This : in out Object);

   ------------------------------------------------------------------------
   -- Scheduled                                                          --
   ------------------------------------------------------------------------
   function Scheduled (This : in Object) return Boolean;

   ------------------------------------------------------------------------
   -- Set_scheduled                                                      --
   ------------------------------------------------------------------------
   procedure Set_scheduled (This : in out Object; Sched : in Boolean := true);

   ------------------------------------------------------------------------
   -- Failures                                                           --
   ------------------------------------------------------------------------
   function Failures (This : in Object) return Natural;

   ------------------------------------------------------------------------
   -- Set_failures                                                       --
   ------------------------------------------------------------------------
   procedure Set_failures (This : in out Object; Failures : in Natural := 0);

   ------------------------------------------------------------------------
   -- Relocate                                                           --
   ------------------------------------------------------------------------
   -- Re-checks the location, just in case the IP range has been moved.
   procedure Relocate (This : in out Object);

   ------------------------------------------------------------------------
   -- Get_country_code                                                   --
   ------------------------------------------------------------------------
   function Get_country_code (This : in Object) 
      return Agpl.Geoip.Country_code;

   ------------------------------------------------------------------------
   -- Clear neighbors                                                    --
   ------------------------------------------------------------------------
   procedure Clear_neighbors (This : in out Object);

   ------------------------------------------------------------------------
   -- Add_neighbor                                                       --
   ------------------------------------------------------------------------
   procedure Add_neighbor (This : in out Object; Neigh : in Object_access);

   ------------------------------------------------------------------------
   -- Make_neighbors                                                     --
   ------------------------------------------------------------------------
   -- Marks both hubs as neighbors, preserving the oldest reach time.
   procedure Make_neighbors (L, R : access Object);
      
   ------------------------------------------------------------------------
   -- Get_neighbors                                                      --
   ------------------------------------------------------------------------
   procedure Get_neighbors (
      This : in Object; Neighbors : out Links_map.Container_type);

   ------------------------------------------------------------------------
   -- Alive_neighbors                                                    --
   ------------------------------------------------------------------------
   -- Says how many alive and with leaves > 0 neighbors a hub has
   function Alive_neighbors (This : in Object) return Natural;

   ------------------------------------------------------------------------
   -- Get_link_duration                                                  --
   ------------------------------------------------------------------------
   -- Says for how many time a hub has been connected to another
   function Get_link_duration (This : in Object; Neigh : in String) 
      return Duration;

private

   use Hub_map;
   use type Links_map.Iterator_type;
   use type Calendar.Time;

   type Object is record
      -- Address in human format (includes port):
      Address : Ustring := Null_ustring;
      -- Leaves:
      Leaves, Max_leaves  : Natural := 0;
      -- Pings done after last query:
      Pings   : Natural := 0;
      -- Currently alive?
      Alive   : Boolean := false;
      -- Answer pending?
      Answer_pending : Boolean := false;
      -- Scheduled for a future check?
      Scheduled      : Boolean := false;
      -- Consecutive failures: 
      Failures       : Natural := 0;

      -- Neighbors:
      Neighbors : Links_map.Container_type;

      -- Location
      Country_code : Agpl.Geoip.Country_code := "??";
      Country_name : Ustring                 := U ("Unknown");

      -- Stats:
      Stat_pings            : Natural := 0;
      Stat_queries          : Natural := 0;
      Stat_answers          : Natural := 0;
      Stat_answers_in_a_row : Natural := 0;
      Stat_first_seen       : Calendar.Time; -- First seen in this "aliveness"
      Stat_last_seen        : Calendar.Time; -- Last seen alive

   end record;

end Aenea.Hub;
