------------------------------------------------------------------------------
--                         ADAGIO - ADALID - AENEA.                         --
--                                                                          --
--                            Copyright (C) 2003                            --
--                                 A. Mosteo.                               --
--                                                                          --
--  Authors: A. Mosteo. (adagio@mosteo.com)                                 --
--                                                                          --
--  If you have any questions in regard to this software, please address    --
--  them to the above email.                                                --
--                                                                          --
--  This program is free software; you can redistribute it and/or modify    --
--  it under the terms of the GNU General Public License as published by    --
--  the Free Software Foundation; either version 2 of the License, or (at   --
--  your option) any later version.                                         --
--                                                                          --
--  This program is distributed in the hope that it will be useful, but     --
--  WITHOUT ANY WARRANTY; without even the implied warranty of              --
--  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU       --
--  General Public License for more details.                                --
--                                                                          --
--  You should have received a copy of the GNU General Public License       --
--  along with this library; if not, write to the Free Software Foundation, --
--  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.          --
--                                                                          --
--  You are not allowed to use any part of this code to develop a program   --
--  whose output would be used to harass or prosecute other users of the    --
--  networks Adagio connects with. All data collected with Adagio or a tool --
--  containing Adagio code about other network users must remain            --
--  confidential and cannot be made public by any mean, nor be used to      --
--  harass or legally prosecute these users.                                --
------------------------------------------------------------------------------
--  $Id: aenea-walker.adb,v 1.13 2004/03/09 23:46:48 Jano Exp $

with Adagio.G2.Packet;
with Adagio.G2.Packet.Parsing;
with Adagio.G2.Packet.Queue;
with Adagio.G2.Transceiver;
with Adagio.Network.Endian;
with Adagio.Socket;

with Aenea.Chronos;
with Aenea.Countries;
with Aenea.DB;
with Aenea.Globals;
with Aenea.Globals.Options;
with Aenea.Gui.Events;
with Aenea.Hub.Report;
with Aenea.Hub.Schedule;
with Aenea.Hub.Topo.Placer_factory;
with Aenea.Net;
with Aenea.Trace;

with Agpl.Safe_file;
use  Agpl;

with Ada.Calendar;
with Ada.Real_time; 
use  Ada.Real_time;
with Ada.Streams.Stream_IO;
use  Ada;

package body Aenea.Walker is

   use type Calendar.Time;
   use Hub.Hub_map;
   package Conv renames Adagio.Network.Endian;

   Hosts_path : constant String := Globals.Data_folder & "hosts.dat";

   -- Null context
   Null_context : Event_queue.Context_type;

   ------------------------------------------------------------------------
   -- Hub collection                                                     --
   ------------------------------------------------------------------------
   Hubs : Hub.Hub_map.Container_type;

   ------------------------------------------------------------------------
   -- Count_hub                                                          --
   ------------------------------------------------------------------------
   -- Will increase Counter.Hubs and do other registrations.
   procedure Count_hub (This : in Hub.Object) is
   begin
      Net.Counter.Add (Hubs => 1);
      Countries.Sum_hub (Hub.Get_country_code (This));
   end Count_hub;

   ------------------------------------------------------------------------
   -- Discount_hub                                                       --
   ------------------------------------------------------------------------
   -- Will decrease Counter.Hubs and do other de-registrations.
   procedure Discount_hub (This : in Hub.Object) is
   begin
      Net.Counter.Add (Hubs => -1);
      Countries.Sum_hub (Hub.Get_country_code (This), -1);
   end Discount_hub;

   ------------------------------------------------------------------------
   -- Get_leaves_from_HS                                                 --
   ------------------------------------------------------------------------
   procedure Get_leaves_from_HS (
      P          : in G2.Packet.Object; 
      Leaves     : out Natural; 
      Max_leaves : out Natural) 
   is
      Data : String := G2.Packet.Payload (P);
   begin
      Leaves := Conv.Convert (
         Conv.To_byte_array (Data (Data'First .. Data'First + 1)),
            G2.Packet.Big_endian (P));
      if Data'Length >= 4 then
         Max_leaves := Conv.Convert (
            Conv.To_byte_array (
               Data (Data'First + 2 .. Data'First + 3)),
            G2.Packet.Big_endian (P));
      else
         Max_leaves := 0;
      end if;
   end Get_leaves_from_HS;

   ------------------------------------------------------------------------
   -- Trace_status                                                       --
   ------------------------------------------------------------------------
   procedure Trace_status is
   begin
      Trace.Log ("Hubs:" & 
         Integer'Image (Net.Counter.Hubs_count) & "; Leaves:" &
         Integer'Image (Net.Counter.Leaves_count) & "; All:" &
         Integer'Image (Length (Hubs)) & "; Pout:" &
         Integer'Image (Net.Outgoing.Length) & "; Ev:" &
         Integer'Image (Event_queue.Length (Net.Events)), 
         Trace.Informative);
   end Trace_status;

   ------------------------------------------------------------------------
   -- Reporter                                                           --
   ------------------------------------------------------------------------
   task Reporter;
   task body Reporter is
      H, L   : Natural := 0;
      NH, NL : Natural;
      Count  : Natural := 0;

      Period   : constant Duration := 1.0;
      Next     : Calendar.Time := Calendar.Clock + Period;
      DBPeriod : Duration := 0.0;
      DBStart  : Calendar.Time := Calendar.Clock;
   begin
      while not Globals.Requested_exit loop
         delay until Next;
         Next := Next + Period;

         NH := Net.Counter.Hubs_count;
         NL := Net.Counter.Leaves_count;
         if NH /= H or else NL /= L then
            Trace_status;
            H := NH; 
            L := NL;
         end if;

         -- Averages
         if Calendar.Clock - DBStart >= Globals.Options.Walk_InsertDelay then
            Count := Count + 1;
            Net.Counter.Push_hubs (NH);
            Net.Counter.Push_leaves (NL);
            if Count = 60 then
               Count := 0;
               Trace.Log ("******* AVG HUBS  :" & 
                  Integer'Image (Net.Counter.Avg_hubs), Trace.Informative);
               Trace.Log ("******* AVG LEAVES:" & 
                  Integer'Image (Net.Counter.Avg_leaves), Trace.Informative);
            end if;
         end if;

         -- DB stuff
         DBPeriod := DBPeriod + 1.0;
         if Calendar.Clock - DBStart >= Globals.Options.Walk_InsertDelay then
            if DBPeriod >= Globals.Options.Walk_InsertPeriod then
               DB.Insert_row (
                  Net.Counter.Avg_hubs,
                  Net.Counter.Avg_leaves,
                  Adder.Count);
               DBPeriod := 0.0;     
            end if;
         end if;
      end loop;
   end Reporter;

   ------------------------------------------------------------------------
   -- Check_queue                                                        --
   ------------------------------------------------------------------------
   procedure Check_queue (C : in Event_queue.Context_access) is
      pragma Unreferenced (C);
      E : Event_queue.Event_type;

      Item : G2.Packet.Queue.Item_type;
   begin
      Trace.Log ("Checking queues...", Trace.Never);
      begin
         loop
            exit when Net.Outgoing.Is_empty and Net.Incoming.Is_empty;

            -- Send outgoing packets
            if not Net.Outgoing.Is_empty then 
               Net.Outgoing.Get (Item);
               Net.Delayer.Request;
               G2.Transceiver.Send (Net.UDPT.all, Item);
            end if;

            -- Process incoming packets
            while not Net.Incoming.Is_empty loop
               Net.Incoming.Get (Item);
               Adder.Process_item (Item);
            end loop;
         end loop;

      exception
         when E : others =>
            Trace.Log (
               "Walker.Check_queue: " & Trace.Report (E), Trace.Error);
      end;

      -- Last, reprogram itself:
      Event_queue.Create (
         Net.Events, E, Clock + Milliseconds (100), 
         Check_queue'Access, Null_context);
   end Check_queue;

   ------------------------------------------------------------------------
   -- Start                                                              --
   ------------------------------------------------------------------------
   -- Creates the first event to process received packets.
   procedure Start is
      E : Event_queue.Event_type;
   begin
      if Globals.Options.Walk_SaveHubs then
         Adder.Restore_hubs;
      end if;
      Event_queue.Create (
         Net.Events, E, 
         Clock + To_time_span (Globals.Options.Walk_SavePeriod), 
         Event_save_hubs'Access, Null_context);
      Event_queue.Create (
         Net.Events, E, Clock, Check_queue'Access, Null_context);
   end Start;

   ------------------------------------------------------------------------
   -- Add_hub                                                            --
   ------------------------------------------------------------------------
   -- Adds a hub to the mesh of known hubs.
   -- It will be immediately queried.
   protected body Adder is
   procedure Add_hub (This : in Hub.Object; Query : in Boolean := true) is
      H : Hub.Object_access;
   begin
      -- Add to mesh if not exists
      if not Find (Hubs, Hub.Id (This)) then
         H := new Hub.Object'(This);
         Insert (Hubs, Hub.Id (H.all), H);
      else
         H := Element (Find (Hubs, Hub.Id (This)));
      end if;
      -- Query it!
      if Query then
         Hub.Query (H.all);
         Hub.Schedule.Create_timeout (H.all);
         -- Schedule a new query/ping for it:
         Hub.Schedule.Reschedule (H.all);
      end if;
   end Add_hub;

      function  Count return Natural is
      begin
         return Length (Hubs);
      end Count;

   ------------------------------------------------------------------------
   -- Save_hubs                                                          --
   ------------------------------------------------------------------------
   procedure Save_hubs is
      use Ada.Streams.Stream_IO;
      I : Iterator_type := First (Hubs);
      H : Hub.Object_access;
      F : File_type;
      S : Stream_access;
      C : Chronos.Object;
   begin 
      Safe_file.Open (F, Out_file, Name => Hosts_path);
      S := Stream (F);
      while I /= Back (Hubs) loop
         H := Element (I);
         Hub.Object'Write (S, H.all);
         I := Succ (Hubs, I);
      end loop;
      Safe_file.Close (F);
      Trace.Log (
         Integer'Image (Length (Hubs)) & " hosts saved in " & 
         Chronos.Image (C), Trace.Informative);
   exception
      when E : others =>
         if Is_open (F) then
            Safe_file.Close (F);
         end if;
         Trace.Log ("Walker.Save_hubs: " & Trace.Report (E), Trace.Error);
   end Save_hubs;

   ------------------------------------------------------------------------
   -- Restore_hubs                                                       --
   ------------------------------------------------------------------------
   procedure Restore_hubs is
      use Ada.Streams.Stream_IO;
      H : Hub.Object;
      F : File_type;
      S : Stream_access;
      C : Chronos.Object;
   begin
      if not Safe_file.Exists_for_reading (Hosts_path) then
         Trace.Log ("No prior hubs found.", Trace.Informative);
         return;
      end if;

      Safe_file.Open (F, In_file, Name => Hosts_path);
      S := Stream (F);
      while not End_of_file (F) loop
         Hub.Object'Read (S, H);
         Hub.Set_scheduled (H, false);
         Hub.Set_alive (H, false);
         Hub.Set_answer_pending (H, false);
         Hub.Set_leaves (H, 0);
         Hub.Kill (H);
         Hub.Relocate (H);
         Add_hub (H, false);
      end loop;
      Safe_file.Close (F);
      Trace.Log (
         Integer'Image (Length (Hubs)) & " hosts restored in " & 
         Chronos.Image (C), Trace.Informative);
   exception
      when E : others =>
         if Is_open (F) then
            Safe_file.Close (F);
         end if;
         Trace.Log ("Walker.Restore_hubs: " & Trace.Report (E), Trace.Error);
   end Restore_hubs;

   ------------------------------------------------------------------------
   -- Report                                                             --
   ------------------------------------------------------------------------
   procedure Report (Data : out Agpl.Http.Server.Sort_handler.Data_set) is
      use Agpl.Http.Server.Sort_handler;
      I : Iterator_type     := First (Hubs);
      H : Hub.Object_access;
   begin
      while I /= Back (Hubs) loop
         H := Element (I);
         declare
            Row : Data_row;
         begin
            if Hub.Alive (H.all) and then Hub.Leaves (H.all) > 0 then
               Hub.Report (H.all, Row);
               Append (Data, Row);
            end if;
         end;
         I := Succ (Hubs, I);
      end loop;
   end Report;

   ------------------------------------------------------------------------
   -- Topogram                                                           --
   ------------------------------------------------------------------------
   function  Topogram (
      Width, Height : in Positive; Kind : in Hub.Topo.Kinds) 
      return Agpl.Bmp.Object
   is
      Data : Hub.Topo.Topo_maps.Container_type;
   begin
      Hub.Topo.Prepare_topogram (Hubs, Data);

      return Hub.Topo.Build_topogram (
         Data, Hub.Topo.Placer_factory.Get (Kind, Width, Height));
   end Topogram;

   ------------------------------------------------------------------------
   -- Process_item                                                       --
   ------------------------------------------------------------------------
   procedure Process_item (Item : in G2.Packet.Queue.Item_type) is
      P     : G2.Packet.Object := Item.Packet;
      I     : Iterator_type;
      H     : Hub.Object_access;
      NH    : Hub.Object;
      NHP   : Hub.Object_access;

      Leaves, Max_leaves : Natural;
   begin
      Trace.Log ("*******", Trace.Debug);
      G2.Packet.Parsing.Trace_tree (P);

      -- Locate or create hub
      I := Find (Hubs, Adagio.Socket.Image (Item.Udp_source));
      if I /= Back (Hubs) then
         H := Element (I);
      else
         H := new Hub.Object;
         Hub.Create (
            H.all, Adagio.Socket.Image (Item.Udp_source), Alive => false);
         Insert (Hubs, Hub.Id (H.all), H);
      end if;

      if Hub.Answer_pending (H.all) then
         Hub.Set_answer_pending (H.all, false);
         Hub.Add_answer (H.all);
         Hub.Set_failures (H.all, 0);
      end if;

      -- Revive a dead hub:
      if not Hub.Alive (H.all) then
         Hub.Set_alive (H.all);
         Hub.Schedule.Reschedule (H.all);
         Hub.Set_failures (H.all, 0);
      end if;
      --------
      -- PO --
      --------
      if G2.Packet.Is_a (Item.Packet, "/PO") then
         -- Nothing to do
         Trace.Log (Hub.Id (H.all) & " is alive.", Trace.Debug);
         Gui.Events.Add ("[<] PO     - " & Hub.Id (H.all), 1);
      ------------
      -- CRAWLA --
      ------------
      elsif G2.Packet.Is_a (Item.Packet, "/CRAWLA") then
         Gui.Events.Add ("[<] CRAWLA - " & Hub.Id (H.all), 1);
         -- Obtain leaves:
         Get_leaves_from_HS (
            G2.Packet.Get_child (Item.Packet, "SELF/HS"), Leaves, Max_leaves);
         Trace.Log (Hub.Id (H.all) & " leaves:" & Leaves'Img &
            "; max:" & Max_leaves'Img);

         if Leaves > 0 and Hub.Leaves (H.all) = 0 then
            Count_hub (H.all);
         elsif
            Leaves = 0 and Hub.Leaves (H.all) > 0 then
            Discount_hub (H.all);
         end if;
         Net.Counter.Add (
            Leaves     => Leaves - Hub.Leaves (H.all),
            Max_leaves => Max_leaves - Hub.Max_leaves (H.all));

         Hub.Set_leaves (H.all, Leaves);
         Hub.Set_max_leaves (H.all, Max_leaves);

         -- Add neighbors:
         declare
            Nodes : G2.Packet.Object_array :=
               G2.Packet.Get_children (Item.Packet, "NH");
         begin
            Hub.Clear_neighbors (H.all);
            for N in Nodes'Range loop
               declare
                  Addr : String := G2.To_address (
                     G2.Packet.Payload (
                        G2.Packet.Get_child (Nodes (N), "NA")),
                     G2.Packet.Big_endian (Item.Packet));
               begin
                  I := Find (Hubs, Addr);
                  -- Add it only if missing:
                  if I = Back (Hubs) then
                     Hub.Create (NH, Addr, Alive => false);
                     Add_hub (NH);
                     -- Locate it to later mark as neighbor
                     NHP := Element (Find (Hubs, Hub.Id (NH)));
                  else
                     NHP := Element (I);
                     -- If neighbor is dead and not queried, query it again:
                     if not Hub.Scheduled (NHP.all) then
                        Hub.Query (NHP.all);
                        Hub.Schedule.Create_timeout (NHP.all);
                        Hub.Schedule.Reschedule (NHP.all);
                     end if;
                  end if; 
                  Hub.Make_neighbors (H, NHP);
               end;
            end loop;
         end;
      else
         Trace.Log ("Unexpected packet: " & G2.Packet.Name (Item.Packet),
            Trace.Warning);
      end if;
   end Process_item;

   ------------------------------------------------------------------------
   -- Check_hub                                                          --
   ------------------------------------------------------------------------
   -- Sends a query or ping for a hub upon event arrival.
   procedure Check_hub (Context : in Event_queue.Context_access) is
      H : Hub.Object_access;
      C : Hub.Schedule.Hub_context := Hub.Schedule.Hub_context (Context.all);
      I : Iterator_type;
   begin
      -- Get the server:
      I := Find (Hubs, S (C.Hub_id));
      if I /= Back (Hubs) then
         H := Element (I);

         Trace.Log ("Checking hub " & S (C.Hub_id), Trace.Debug);

         Hub.Set_scheduled (H.all, false);
         Gui.Events.Add ("[*] Chk    - " & Hub.Id (H.all), 2);
     
         if Hub.Alive (H.all) and then 
            Hub.Answer_pending (H.all) and then
            Hub.Failures (H.all) = Globals.Options.Walk_Failures
         then -- Shouldn't happen, the timeout must have checked this before.
            Discount_hub (H.all);
            Hub.Kill (H.all);
            Trace.Log("Pending found in Check_hub (1)", Trace.Warning);
         elsif Hub.Alive (H.all) and then Hub.Leaves (H.all) > 0 then
            -- Increase failures if needed:
            if Hub.Answer_pending (H.all) then
               Hub.Set_failures (H.all, Hub.Failures (H.all) + 1);
               Trace.Log("Pending found in Check_hub (2)", Trace.Warning);
            end if;
            -- Send a query or ping:
            case C.Msg_type is
               when Hub.Schedule.Ping =>
                  Hub.Ping (H.all);
                  Hub.Schedule.Create_timeout (H.all);
               when Hub.Schedule.Query =>
                  Hub.Query (H.all);
                  Hub.Schedule.Create_timeout (H.all);
               when others =>
                  raise Unimplemented;
            end case;
            --Trace_status;

            Hub.Schedule.Reschedule (H.all);
         end if;
      end if;
   end Check_hub;

   ------------------------------------------------------------------------
   -- Check_timeout                                                      --
   ------------------------------------------------------------------------
   -- Checks if there are answers pending. In that case, kills the hub.
   procedure Check_timeout (Context : in Event_queue.Context_access) is
      H : Hub.Object_access;
      C : Hub.Schedule.Hub_context := Hub.Schedule.Hub_context (Context.all);
      I : Iterator_type;
   begin
      -- Get the server:
      I := Find (Hubs, S (C.Hub_id));
      if I /= Back (Hubs) then
         H := Element (I);

         Gui.Events.Add ("[*] T-out  - " & Hub.Id (H.all), 1);
         Trace.Log ("Checking timeout for hub " & S (C.Hub_id), Trace.Debug);

         if not Hub.Answer_pending (H.all) then
            return; -- <-- EARLY EXIT IF NO ANSWER IS PENDING
         end if;

         if Hub.Alive (H.all) and then 
            Hub.Failures (H.all) = Globals.Options.Walk_Failures
         then
            Discount_hub (H.all);
            Hub.Kill (H.all);
         elsif Hub.Alive (H.all) and then Hub.Leaves (H.all) > 0 then
            -- Increase failures:
            Hub.Set_failures (H.all, Hub.Failures (H.all) + 1);
            Gui.Events.Add ("[*] Tout+1 - " & Hub.Id (H.all), 2);
         end if;

         Hub.Set_answer_pending (H.all, false);
      end if;
   end Check_timeout;

   end Adder;

   ------------------------------------------------------------------------
   -- Check_hub                                                          --
   ------------------------------------------------------------------------
   -- Sends a query or ping for a hub upon event arrival.
   procedure Check_hub (Context : in Event_queue.Context_access) is
   begin
      Adder.Check_hub (Context);
   end Check_hub;

   ------------------------------------------------------------------------
   -- Check_timeout                                                      --
   ------------------------------------------------------------------------
   -- Checks timeouts for a hub.
   -- May kill the hub if rules say that.
   -- The hub will remaing scheduled, the schedule is removed in Check_hub
   procedure Check_timeout (Context : in Event_queue.Context_access) is
   begin
      Adder.Check_timeout (Context);
   end Check_timeout;
   
   procedure Add_hub (This : in Hub.Object; Query : in Boolean := true) is
   begin
      Adder.Add_hub (This, Query);
   end Add_hub;

   ------------------------------------------------------------------------
   -- Event_save_hubs                                                    --
   ------------------------------------------------------------------------
   procedure Event_save_hubs (Context : in Event_queue.Context_access) is
      E : Event_queue.Event_type;
      pragma Unreferenced (Context);
   begin
      if Globals.Options.Walk_SaveHubs then
         Event_queue.Create (
            Net.Events, E, 
            Clock + To_time_span (Globals.Options.Walk_SavePeriod), 
            Event_save_hubs'Access, Null_context);
         Adder.Save_hubs;
      end if;
   end Event_save_hubs;

   ------------
   -- Report --
   ------------
   procedure Report (Data : out Agpl.Http.Server.Sort_handler.Data_set) is
   begin
      Adder.Report (Data);
   end Report;

end Aenea.Walker;
