------------------------------------------------------------------------------
--                         ADAGIO - ADALID - AENEA.                         --
--                                                                          --
--                            Copyright (C) 2003                            --
--                                 A. Mosteo.                               --
--                                                                          --
--  Authors: A. Mosteo. (adagio@mosteo.com)                                 --
--                                                                          --
--  If you have any questions in regard to this software, please address    --
--  them to the above email.                                                --
--                                                                          --
--  This program is free software; you can redistribute it and/or modify    --
--  it under the terms of the GNU General Public License as published by    --
--  the Free Software Foundation; either version 2 of the License, or (at   --
--  your option) any later version.                                         --
--                                                                          --
--  This program is distributed in the hope that it will be useful, but     --
--  WITHOUT ANY WARRANTY; without even the implied warranty of              --
--  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU       --
--  General Public License for more details.                                --
--                                                                          --
--  You should have received a copy of the GNU General Public License       --
--  along with this library; if not, write to the Free Software Foundation, --
--  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.          --
--                                                                          --
--  You are not allowed to use any part of this code to develop a program   --
--  whose output would be used to harass or prosecute other users of the    --
--  networks Adagio connects with. All data collected with Adagio or a tool --
--  containing Adagio code about other network users must remain            --
--  confidential and cannot be made public by any mean, nor be used to      --
--  harass or legally prosecute these users.                                --
------------------------------------------------------------------------------
--  $Id: agpl-streams-deflate.adb,v 1.1 2004/03/29 19:13:33 Jano Exp $

--  Uses a internal circular buffer

package body Agpl.Streams.Deflate is

   -- Helper always ready function, will return Natural'Last:
   function Always_available (
      This : access Ada.Streams.Root_stream_type'Class) return Natural
   is
      pragma Unreferenced (This);
   begin
      return Natural'Last;
   end Always_available;

   ------------------------------------------------------------------------
   -- Create                                                             --
   ------------------------------------------------------------------------
   -- The percent free says when Ready_for_writing will return true, according
   -- to the internal buffer filling.
   procedure Create (
      Stream       :    out Stream_type;
      From         : access Ada.Streams.Root_stream_type'Class; 
      Avail_read   : in     Available_function := Always_available'Access;
      Avail_write  : in     Available_function := Always_available'Access;
      Buffer       : in     Positive := 4096;
      Percent_free : in     Natural  := 20)
   is
   begin
      Stream.Back := Agpl.Streams.Stream_acces (From);
      Stream.Avail_read  := Avail_read;
      Stream.Avail_write := Avail_write;
      Stream.Buffer_size := Buffer;
      Stream.Percent_free := Percent_free;
      Stream.Buf_in       := new Circular.Stream_type (Buffer);
      Stream.Buf_out      := new Circular.Stream_type (Buffer);

      Zlib.Streams.Create (
         Zin, 
         Zlib.Streams.Out_mode,
         Streams.Buf_in,
         Back_compressed => false);
      Zlib.Streams.Create (
         Zout, 
         Zlib.Streams.Out_mode,
         Streams.Buf_out,
         Back_compressed => true);
   end Create;

   ------------------------------------------------------------------------
   -- Read                                                               --
   ------------------------------------------------------------------------
   procedure Read(
      Stream : in out Stream_type;
      Item   : out Ada.Streams.Stream_Element_Array;
      Last   : out Ada.Streams.Stream_Element_Offset)
   is
   begin
      Attempt_read (Stream);
      Circular.Read (Stream.Buf_in.all, Item, Last);
   end Read;

   ------------------------------------------------------------------------
   -- Write                                                              --
   ------------------------------------------------------------------------
	procedure Write(
      Stream : in out Stream_type;
      Item   : in Ada.Streams.Stream_Element_Array) 
   is
      Pos    : Ada.Streams.Stream_element_offset := Item'First;
      Next   : Ada.Streams.Stream_element_offset;
      Free   : constant Natural := 
         Stream.Buffer_size * Stream.Percent_free / 100;
   begin
      Attempt_write (Stream);
      if Stream.Pending_write /= null then
         raise Constraint_error;
      end if;

      -- Write while possible, then keep in buffer the remaining data:
      while Pos <= Item'Last and then
         Circular.Available_write (Stream.Buf_out.all) >= Free 
      loop
         Next := Ada.Streams.Stream_element_offset'Min (
            Item'Last - Pos + 1,
            Ada.Streams.Stream_element_offset (Free) / 2);
         Zlib.Streams.Write (Zout, Item (Pos .. Pos + Next - 1);
         Pos := Pos + Next;
      end loop;

      if Pos <= Item'Last then
         Stream.Pending_write := new Ada.Streams.Stream_element_array'(
            Item (Pos .. Item'Last);
      end if;

   end Write;

   ------------------------------------------------------------------------
   -- Available_read                                                     --
   ------------------------------------------------------------------------
   -- Says how many data is ready to be read
   function Available_read (This : in Stream_type) return Natural;

   ------------------------------------------------------------------------
   -- Ready_for_writing                                                  --
   ------------------------------------------------------------------------
   function Ready_for_writing (This : in Stream_type) return Boolean;

private

   type Stream_finalizer (Parent : access Stream_type) is new
      Finalization.Limited_controlled with null record;

   procedure Finalize (This : in out Stream_finalizer);

   type Stream_type
   is new Ada.Streams.Root_stream_type with record
      Buf_in  : Agpl.Streams.Circular.Stream_access;
      Buf_out : Agpl.Streams.Circular.Stream_access;

      Percent_free : Natural;
      Buffer_size  : Natural;

      Avail_read,
      Avail_write  : Available_function;

      -- Buffers with data pending to be read/write to the CStreams
      Pending_read  : Agpl.Streams.Stream_element_array_access;
      Pending_write : Agpl.Streams.Stream_element_array_access;

      -- Finalizer:
      Finalize : Stream_finalizer (Stream_type'Access);
   end record;

   ------------------------------------------------------------------------
   -- Attempt_read                                                       --
   ------------------------------------------------------------------------
   -- Read as many data as possible restricted by availabilities.
   procedure Attempt_read (Stream : in Stream_type);

   ------------------------------------------------------------------------
   -- Attempt_write                                                      --
   ------------------------------------------------------------------------
   -- Write as many data as possible restricted by availabilities.
   procedure Attempt_write (Stream : in Stream_type);

end Agpl.Streams.Deflate;
